#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2019/11/11 11:11
# @Author  : niuliangtao
# @Site    :
# @Software: PyCharm
import os

import demjson
import requests


class AccessToken:
    def __init__(self, api_key=None, secret_key=None, access_token=None, refresh_token=None, secret_dir='~/.secret'):
        self.secret = None
        secret_dir = secret_dir.replace("~", os.environ['HOME'])
        self.secret_path = '{}/.baidu_drive'.format(secret_dir)
        self.read()

        if api_key is not None:
            self.secret['token']['client_id'] = api_key
        if secret_key is not None:
            self.secret['token']['client_secret'] = secret_key

        if refresh_token is not None:
            self.secret['token']['refresh_token'] = refresh_token
        if access_token is not None:
            self.secret['token']['access_token'] = access_token

        self.write()

    def get_authorization_code(self):
        url1 = 'https://openapi.baidu.com/oauth/2.0/authorize?response_type=code&client_id={}' \
               '&redirect_uri={}&scope=netdisk'.format(self.secret['token']['client_id'], 'oob')

        print(url1)

        self.secret['token']['authorization_code'] = input("点击网页授权并输入authorization_code：")
        self.write()

    def init_refresh_token(self):
        self.get_authorization_code()
        url2 = 'https://openapi.baidu.com/oauth/2.0/token?grant_type=authorization_code&code={}&client_id={}&' \
               'client_secret={}&scope=basic,netdisk&redirect_uri=oob' \
            .format(self.secret['token']['authorization_code'],
                    self.secret['token']['client_id'],
                    self.secret['token']['client_secret'])
        print(url2)

        res = requests.post(url2)
        tokens = demjson.decode(res.text)
        print(tokens)
        self.secret['token']['access_token'] = tokens['access_token']
        self.secret['token']['refresh_token'] = tokens['refresh_token']
        self.write()

    def refresh_access_token(self):
        url3 = 'https://openapi.baidu.com/oauth/2.0/token?grant_type=refresh_token&refresh_token={}&client_id={}&' \
               'scope=basic+netdisk&client_secret={}'.format(self.secret['token']['refresh_token'],
                                                             self.secret['token']['client_id'],
                                                             self.secret['token']['client_secret'])

        try:
            print(url3)
            res3 = requests.post(url3)

            token_new = demjson.decode(res3.text)
            print(token_new)
            self.secret['token']['access_token'] = token_new['access_token']
            self.secret['token']['refresh_token'] = token_new['refresh_token']
        except Exception as e:
            print(e)
        self.write()

    def get_access_token(self):
        if self.secret['token']['access_token'] is not None:
            return self.secret['token']['access_token']
        elif self.secret['token']['refresh_token'] is not None:
            self.refresh_access_token()

        else:
            self.init_refresh_token()
            self.refresh_access_token()

        self.write()
        return self.secret['token']['access_token']

    def read(self):
        try:
            self.secret = demjson.decode(open(self.secret_path).read())
        except Exception as e:
            print("read error ,init {}".format(e))

        if self.secret is None:
            self.secret = {'token': {}}
        if 'token' not in self.secret.keys() or not isinstance(self.secret['token'], dict):
            self.secret['token'] = {}

        for key in ['api_key', 'secret_key', 'access_token', 'refresh_token', 'authorization_code']:
            if key not in self.secret['token'].keys():
                self.secret['token'][key] = ""

        return self.secret

    def write(self):
        secret_dir = os.path.dirname(self.secret_path)

        if not os.path.exists(secret_dir):
            os.mkdir(secret_dir)

        with open(self.secret_path, 'w')as f:
            f.write(demjson.encode(self.secret))



import asyncio
import hashlib
import os
import random
import struct
from asyncio import DatagramProtocol
from base64 import b64encode
from ipaddress import IPv4Address

from expiringdict import ExpiringDict

from notetool.tool.log import log
from .bencode import BTFailure, bdecode, bencode
from .node import Node
from .routing import RoutingTable
from .utils import digest

"""
Some code taken from
https://github.com/bmuller/rpcudp/blob/master/rpcudp/protocol.py
Has same license as kademlia, see __init__.py

https://www.cnblogs.com/LittleHann/p/6180296.html

DHT网络的其中一种协议实现(Kademlia)
要加入一个DHT网络，需要首先知道这个网络中的任意一个节点。如何获得这个节点？在一些开源的P2P软件中，会提供一些节点地址
****************************************************************************************************************
* 1.ping           用于确定某个节点是否在线。这个请求主要用于辅助路由表的更新
* 2.find_node      用于查找某个节点，以获得其地址信息。
* 3.get_peer       通过资源的infohash获得资源对应的peer列表。
* 4.announce_peer  通知其他节点自己开始下载某个资源,announce_peer中会携带get_peer回应消息里的token。 
****************************************************************************************************************
"""

MIN_ID = 0
MAX_ID = 2 ** 160


class ForgetfulPeerStorage:
    logger = log(__name__)

    def __init__(self, ttl=3600):
        self._ttl = ttl
        self._data = ExpiringDict(max_age_seconds=ttl, max_len=2000)
        self.data = ExpiringDict(max_age_seconds=ttl, max_len=2000)

    def get_peers(self, info_hash):
        if info_hash not in self._data:
            return []
        return self._data[info_hash].keys()

    def insert_peer(self, info_hash, peer):
        self.data[info_hash][peer] = None
        self.data[info_hash] = self.data[info_hash]


class ForgetfulTokenStorage:
    def __init__(self, ttl=600):
        self._data = ExpiringDict(max_age_seconds=ttl, max_len=2000)

    def get_token(self, sender, id, info_hash):
        token = bytes([random.randint(0, 255) for _ in range(16)])
        self._data[token] = (sender[0], id, info_hash)
        return token

    def verify_token(self, sender, id, info_hash, token):
        if self._data.get(token) == (sender[0], id, info_hash):
            del self._data[token]
            return True
        return False


class KRPCProtocol(DatagramProtocol):
    logger = log(__name__)

    def __init__(self,
                 source_node: Node = None,
                 peer_storage: ForgetfulPeerStorage = None,
                 token_storage: ForgetfulTokenStorage = None,
                 ksize: int = 8,
                 wait_timeout=60,
                 buckets=None, ):
        self.router = RoutingTable(self, ksize, source_node, buckets=buckets)
        self.peer_storage = peer_storage or ForgetfulPeerStorage()
        self.token_storage = token_storage or ForgetfulTokenStorage()
        self.source_node = source_node or Node(digest(random.getrandbits(255)))
        self._wait_timeout = wait_timeout
        self._outstanding = {}
        self.transport = None

    def connection_made(self, transport):
        """Called when a connection is made.

        The argument is the transport representing the pipe connection.
        To receive data, wait for data_received() calls.
        When the connection is closed, connection_lost() is called.
        """
        self.transport = transport

    def datagram_received(self, data, addr):
        """Called when some datagram is received."""

        self.logger.debug("received datagram from %s", addr)
        try:
            data = bdecode(data)
        except (BTFailure, ValueError, KeyError):
            self.logger.info("Failed to decode message")
            return

        if not isinstance(data, dict):
            return

        query_type = data.get(b"y")
        transaction_id = data.get(b"t")
        if not transaction_id:
            return

        if query_type == b"q":  # query
            func_name, args = data.get(b"q"), data.get(b"a")
            if func_name and isinstance(args, dict):
                asyncio.ensure_future(self.handle_request(transaction_id, func_name, args, addr))
        elif query_type == b"r":  # response
            args = data.get(b"r")
            if isinstance(args, dict):
                asyncio.ensure_future(self.handle_response(transaction_id, args, addr))
        elif query_type == b"e":  # error
            args = data.get(b"e")
            if isinstance(args, list):
                self.handle_error(transaction_id, args, addr)
            return
        else:
            return

    async def handle_request(self, transaction_id, func_name, args, addr):
        """
        q 表示请求(请求Queries): q类型的消息它包含 2 个附加的关键字 q 和 a
            关键字 q: 是字符串类型，包含了请求的方法名字(get_peers/announce_peer/ping/find_node)
            关键字 a: 一个字典类型包含了请求所附加的参数(info_hash/id..)
        :param transaction_id:
        :param func_name: 对应关键词 q
        :param args:      对应关键词 a
        :param addr:
        :return:
        """
        func = getattr(self, "rpc_{}".format(func_name.decode('utf-8')), None)
        if func is None or not callable(func):
            msg_args = (self.__class__.__name__, func_name)
            self.logger.info("%s has no callable method " "rpc_%s; ignoring request", *msg_args)
            return

        if not asyncio.iscoroutinefunction(func):
            func = asyncio.coroutine(func)
        args = {k.decode("utf-8"): v for (k, v) in args.items()}
        response = await func(addr, **args)
        if response is not None:
            self.logger.debug("sending response %s for msg id %s to %s", response, b64encode(transaction_id), addr, )
            tx_data = bencode({b"y": b"r", b"r": response, })
            self.transport.sendto(tx_data, addr)

    async def handle_response(self, transaction_id, args, addr):
        """
        r 表示回复(回复 Responses): 包含了返回的值。发送回复消息是在正确解析了请求消息的基础上完成的，包含了一个附加的关键字 r。关键字 r 是字典类型
            id: peer节点id号或者下一跳DHT节点
            nodes": ""
            token: token
        :param transaction_id:
        :param args:
        :param addr:
        :return:
        """
        msg_args = (b64encode(transaction_id), addr)
        if transaction_id not in self._outstanding:
            self.logger.info("received unknown message %s " "from %s; ignoring", *msg_args)
            return
        self.logger.debug("received response %s for message " "id %s from %s", args, *msg_args)
        future, timeout = self._outstanding[transaction_id]
        timeout.cancel()
        if not future.cancelled():
            future.set_result((True, args))
        del self._outstanding[transaction_id]

    def handle_error(self, transaction_id, args, addr):
        """
        3) e 表示错误(错误 Errors): 包含一个附加的关键字 e，关键字 e 是列表类型
            3.1) 第一个元素是数字类型，表明了错误码，当一个请求不能解析或出错时，错误包将被发送。下表描述了可能出现的错误码
                201: 一般错误
                202: 服务错误
                203: 协议错误，比如不规范的包，无效的参数，或者错误的 toke
                204: 未知方法
            3.2) 第二个元素是字符串类型，表明了错误信息
        :return:
        """
        pass

    def _timeout(self, transaction_id):
        args = (b64encode(transaction_id), self._wait_timeout)
        self.logger.info("Did not received reply for msg " "id %s within %i seconds", *args)
        future = self._outstanding[transaction_id][0]
        if not future.cancelled():
            future.set_result((False, None))
        del self._outstanding[transaction_id]

    def get_refresh_ids(self):
        """
        Get ids to search for to keep old buckets up to date.
        """
        ids = []
        for bucket in self.router.lonely_buckets():
            rid = random.randint(*bucket.range).to_bytes(20, byteorder="big")
            ids.append(rid)
        return ids

    @staticmethod
    def is_valid_node_id(node):
        return MIN_ID < node.long_id < MAX_ID

    def rpc_ping(self, sender, id):
        source = Node(id, sender[0], sender[1])
        if not self.is_valid_node_id(source):
            return
        self.welcome_if_new(source)
        return {b"id": id}

    def rpc_announce_peer(self, sender, id, info_hash, port, token, name=None, implied_port=None, seed=None, ):
        source = Node(id, sender[0], sender[1])
        if not self.is_valid_node_id(source):
            return

        self.welcome_if_new(source)
        if self.token_storage.verify_token(sender[0], id, info_hash, token):
            if implied_port:
                port = sender[1]
            self.logger.debug("got an announce_peer request from %s, storing '%s'", sender, info_hash.hex(), )
            self.peer_storage.insert_peer((sender[0], port))
        else:
            self.logger.debug("Invalid token from %s", sender)
        return {b"id": id}

    def rpc_find_node(self, sender, id, target, want="n4", token=None):
        self.logger.info("finding neighbors of %i in local table", int(id.hex(), 16))
        source = Node(id, sender[0], sender[1])
        if not self.is_valid_node_id(source):
            return

        self.welcome_if_new(source)
        node = Node(target)
        if not self.is_valid_node_id(node):
            return
        neighbors = self.router.find_neighbors(node, exclude=source)
        data = {b"id": id, b"nodes": b"".join([n.packed for n in neighbors])}
        if token:
            data[b"token"] = token
        return data

    def rpc_get_peers(self, sender, id, info_hash, want="n4", noseed=0, scrape=0, bs=None):
        source = Node(id, sender[0], sender[1])
        if not self.is_valid_node_id(source):
            return
        self.welcome_if_new(source)
        peers = self.peer_storage.get_peers(info_hash)
        token = self.token_storage.get_token(sender, id, info_hash)
        if not peers:
            return self.rpc_find_node(sender, id, info_hash, token=token)

        return {
            b"id": id,
            b"token": token,
            b"values": [
                IPv4Address(peer[0]).packed + struct.pack("!H", peer[1])
                for peer in peers
            ],
        }

    async def call_ping(self, node_to_ask):
        """
        ping: 检测节点是否可达，请求包含一个参数id，代表该节点的nodeID。对应的回复也应该包含回复者的nodeID
            Query = {"t":"aa", "y":"q", "q":"ping", "a":{"id":"abcdefghij0123456789"}}
            bencoded = d1:ad2:id20:abcdefghij0123456789e1:q4:ping1:t2:aa1:y1:qe

            Response = {"t":"aa", "y":"r", "r": {"id":"mnopqrstuvwxyz123456"}}
            bencoded = d1:rd2:id20:mnopqrstuvwxyz123456e1:t2:aa1:y1:re
        :param node_to_ask:
        :return:
        """
        address = (node_to_ask.ip, node_to_ask.port)
        result = await self.ping(address, {b"id": self.source_node.id})
        return self.handle_call_response(result, node_to_ask)

    async def call_find_node(self, node_to_ask, node_to_find):
        """
        2. find_node: find_node 被用来查找给定 ID 的DHT节点的联系信息，该请求包含两个参数id(代表该节点的nodeID)和target。回复中应该包含被请求节点的路由表中距离target最接近的K个nodeID以及对应的nodeINFO
            # "id" containing the node ID of the querying node, and "target" containing the ID of the node sought by the queryer.
            Query = {"t":"aa", "y":"q", "q":"find_node", "a": {"id":"abcdefghij0123456789", "target":"mnopqrstuvwxyz123456"}}
            bencoded = d1:ad2:id20:abcdefghij01234567896:target20:mnopqrstuvwxyz123456e1:q9:find_node1:t2:aa1:y1:qe

            Response = {"t":"aa", "y":"r", "r": {"id":"0123456789abcdefghij", "nodes": "def456..."}}
            bencoded = d1:rd2:id20:0123456789abcdefghij5:nodes9:def456...e1:t2:aa1:y1:re
        :param node_to_ask:
        :param node_to_find:
        :return:
        """
        address = (node_to_ask.ip, node_to_ask.port)
        result = await self.find_node(address, {b"id": self.source_node.id, b"target": node_to_find.id})
        return self.handle_call_response(result, node_to_ask)

    async def call_get_peers(self, node_to_ask: Node, node_to_find):
        """
        get_peers 请求包含 2 个参数
            1) id请求节点ID，
            2) info_hash代表torrent文件的infohash，infohash为种子文件的SHA1哈希值，也就是磁力链接的btih值
        2. response get_peer:
            1) 如果被请求的节点有对应 info_hash 的 peers，他将返回一个关键字 values，这是一个列表类型的字符串。
               每一个字符串包含了 "CompactIP-address/portinfo" 格式的 peers 信息(即对应的机器ip/port信息)(peer的info信息和DHT节点的info信息是一样的)
            2) 如果被请求的节点没有这个 infohash 的 peers，那么他将返回关键字 nodes
                (需要注意的是，如果该节点没有对应的infohash信息，而只是返回了nodes，则请求方会认为该节点是一个"可疑节点"，则会从自己的路由表K捅中删除该节点)，
               这个关键字包含了被请求节点的路由表中离 info_hash 最近的 K 个节点(我这里没有该节点，去别的节点试试运气)，使用 "Compactnodeinfo" 格式回复。
            在这两种情况下，关键字 token 都将被返回。token 关键字在今后的 annouce_peer 请求中必须要携带。token 是一个短的二进制字符串。

        get_peers Query = {"t":"aa", "y":"q", "q":"get_peers", "a": {"id":"abcdefghij0123456789", "info_hash":"mnopqrstuvwxyz123456"}}
        bencoded = d1:ad2:id20:abcdefghij01234567899:info_hash20:mnopqrstuvwxyz123456e1:q9:get_peers1:t2:aa1:y1:qe

        Response with peers = {"t":"aa", "y":"r", "r": {"id":"abcdefghij0123456789", "token":"aoeusnth", "values": ["axje.u", "idhtnm"]}}
        bencoded = d1:rd2:id20:abcdefghij01234567895:token8:aoeusnth6:valuesl6:axje.u6:idhtnmee1:t2:aa1:y1:re

        Response with closest nodes = {"t":"aa", "y":"r", "r": {"id":"abcdefghij0123456789", "token":"aoeusnth", "nodes": "def456..."}}
        bencoded = d1:rd2:id20:abcdefghij01234567895:nodes9:def456...5:token8:aoeusnthe1:t2:aa1:y1:re
        :param node_to_ask:
        :param node_to_find:
        :return:
        """
        address = (node_to_ask.ip, node_to_ask.port)
        result = await self.get_peers(address, {b"id": self.source_node.id, b"info_hash": node_to_find.id})
        return self.handle_call_response(result, node_to_ask)

    async def call_announce_peer(self, node_to_ask, key, value):  # TODO
        """
        announce_peer: 这个请求用来表明发出 announce_peer 请求的节点，正在某个端口下载 torrent 文件
        announce_peer 包含 4 个参数
            1) 第一个参数是 id: 包含了请求节点的 ID
            2) 第二个参数是 info_hash: 包含了 torrent 文件的 infohash
            3) 第三个参数是 port: 包含了整型的端口号，表明 peer 在哪个端口下载
            4) 第四个参数数是 token: 这是在之前的 get_peers 请求中收到的回复中包含的。
        收到 announce_peer 请求的节点必须检查这个 token 与之前我们回复给这个节点 get_peers 的 token 是否相同
        (也就说，所有下载者/发布者都要参与检测新加入的发布者是否伪造了该资源，但是这个机制有一个问题，如果最开始的那个发布者就伪造，则整条链路都是一个伪造的错的资源infohash信息了)
        如果相同，那么被请求的节点将记录发送 announce_peer 节点的 IP 和请求中包含的 port 端口号在 peer 联系信息中对应的 infohash 下，
        这意味着一个一个事实: 当前这个资源有一个新的peer提供者了，下一次有其他节点希望或者这个资源的时候，会把这个新的(前一次请求下载资源的节点)也当作一个peer返回给请求者，
        这样，资源的提供者就越来越多，资源共享速度就越来越快。
        一个peer正在下载某个资源，意味着该peer有能够访问到该资源的渠道，且该peer本地是有这份资源的全部或部分拷贝的，它需要向DHT网络广播announce消息，告诉其他节点这个资源的下载地址


        announce_peers Query = {"t":"aa", "y":"q", "q":"announce_peer", "a": {"id":"abcdefghij0123456789", "implied_port": 1, "info_hash":"mnopqrstuvwxyz123456", "port": 6881, "token": "aoeusnth"}}
        bencoded = d1:ad2:id20:abcdefghij01234567899:info_hash20:mnopqrstuvwxyz1234564:porti6881e5:token8:aoeusnthe1:q13:announce_peer1:t2:aa1:y1:qe

        Response = {"t":"aa", "y":"r", "r": {"id":"mnopqrstuvwxyz123456"}}
        bencoded = d1:rd2:id20:mnopqrstuvwxyz123456e1:t2:aa1:y1:re

        :param node_to_ask:
        :param key:
        :param value:
        :return:
        """
        address = (node_to_ask.ip, node_to_ask.port)
        result = await self.store(address, self.source_node.id, key, value)
        return self.handle_call_response(result, node_to_ask)

    def welcome_if_new(self, node):
        """
        Given a new node, send it all the keys/values it should be storing,
        then add it to the routing table.

        @param node: A new node that just joined (or that we just found out
        about).

        Process:
        For each key in storage, get k closest nodes.  If newnode is closer
        than the furtherst in that list, and the node for this server
        is closer than the closest in that list, then store the key/value
        on the new node (per section 2.5 of the paper)
        """
        if not self.router.is_new_node(node):
            return

        self.logger.info("never seen %s before, adding to router", node)
        self.router.add_contact(node)

    def handle_call_response(self, result, node):
        """
        If we get a response, add the node to the routing table.  If
        we get no response, make sure it's removed from the routing table.
        """
        if not result[0]:
            self.logger.info("no response from %s, removing from router", node)
            self.router.remove_contact(node)
            return result

        self.logger.info("got successful response from %s", node)
        self.welcome_if_new(node)
        return result

    @staticmethod
    def generate_token():
        return bytes([random.randint(0, 255) for _ in range(16)])

    def __getattr__(self, name):
        """
        If name begins with "_" or "rpc_", returns the value of
        the attribute in question as normal.
        Otherwise, returns the value as normal *if* the attribute
        exists, but does *not* raise AttributeError if it doesn't.
        Instead, returns a closure, func, which takes an argument
        "address" and additional arbitrary args (but not kwargs).
        func attempts to call a remote method "rpc_{name}",
        passing those args, on a node reachable at address.
        """
        if name.startswith("_") or name.startswith("rpc_"):
            return getattr(super(), name)

        try:
            return getattr(super(), name)
        except AttributeError:
            pass

        def func(address, args):
            transaction_id = hashlib.sha1(os.urandom(32)).digest()
            txdata = bencode({b"y": b"q", b"t": transaction_id, b"a": args, b"q": name.encode("utf-8"), })
            self.logger.debug("calling remote function %s on %s (msgid %s)", name, address, b64encode(transaction_id), )
            self.transport.sendto(txdata, address)

            loop = asyncio.get_event_loop()
            if hasattr(loop, "create_future"):
                future = loop.create_future()
            else:
                future = asyncio.Future()
            timeout = loop.call_later(self._wait_timeout, self._timeout, transaction_id)
            self._outstanding[transaction_id] = (future, timeout)
            return future

        return func



import base64
import binascii
from pathlib import Path
from urllib.parse import parse_qs, urlparse, quote

from notetool.tool.log import log
from .bencode import bdecode, bencode

logger = log(__name__)


class Magnet:
    """
    magnet: 协议名。
    xt: exact topic 的缩写，表示资源定位点。BTIH（BitTorrent Info Hash）表示哈希方法名，这里还可以使用 SHA1 和 MD5。
        这个值是文件的标识符，是不可缺少的。
        一般来讲，一个磁力链接只需要上面两个参数即可找到唯一对应的资源。也有其他的可选参数提供更加详细的信息。
    dn: display name 的缩写，表示向用户显示的文件名。
    tr: tracker 的缩写，表示 tracker 服务器的地址。
    kt: 关键字，更笼统的搜索，指定搜索关键字而不是特定文件。
    mt: 文件列表，链接到一个包含磁力链接的元文件 (MAGMA - MAGnet MAnifest）。
    """

    def __init__(self, magnet_link):
        self.name = None
        self.trackers = []
        self.infohash = None

        try:
            self._parse_url(magnet_link)
        except Exception as e:
            logger.error("parse error {}".format(e))

    def _parse_url(self, magnet_link):
        url = urlparse(magnet_link)
        url_query = parse_qs(url.query)
        infohash = url_query["xt"][0].split(":")[2]

        if len(infohash) == 40:
            self.infohash = binascii.unhexlify(infohash)
        elif len(infohash) == 32:
            self.infohash = base64.b32decode(infohash)
        else:
            raise Exception("Unable to parse infohash")

        self.trackers = url_query.get("tr", [])
        name = url_query.get("dn")
        if name:
            self.name = name[0]
        else:
            self.name = infohash  # binascii.hexlify(infohash).decode()

        # TODO: better stripping
        self.name = self.name.strip(".").replace("/", "").replace("\\", "").replace(":", "")


class Torrent:
    def __init__(self, magnet: Magnet = None, data=None, cache_folder=None):
        self.length = 0
        self.name = ''
        self.piece_length = 0
        self.pieces = []
        self.data = data

        self.magnet = magnet
        self._cache_folder = cache_folder
        if self.data:
            self._parse_data()

    def __str__(self):
        return "{}  {}  {}".format(self.name, self.length, self.piece_length)

    def _cache_path(self):
        if self._cache_folder and self.magnet is not None:
            filename = binascii.hexlify(self.magnet.infohash).decode("utf-8")
            return Path(self._cache_folder) / Path(filename[:2]) / Path(filename[2:4]) / Path(filename)

    def _parse_data(self):
        try:
            self.name = quote(self.data[b'info'][b'name'])
            self.length = self.data[b'info'][b'length']
            self.piece_length = self.data[b'info'][b'piece length']
        except Exception as e:
            logger.error("parse error {}".format(e))

    def update_date(self, data):
        self.from_data(data)
        return self

    def from_data(self, data, decode=False):
        if decode:
            self.data = bdecode(data)
        else:
            self.data = data

        self._parse_data()
        return self

    def from_file(self, path):
        path = Path(path)
        if path and path.exists():
            data = Path(path).read_bytes()
            self.from_data(data, decode=True)
        return self

    def to_file(self, path) -> bool:
        if path is None:
            return False
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_bytes(bencode(self.data))

        return True

    def read_cache(self):
        path = self._cache_path()
        if path and path.exists():
            self.from_file(path)
        if self.not_empty():
            logger.info("We had a cache at {}!s".format(path))

    def save_cache(self) -> bool:
        return self.to_file(self._cache_path())

    def empty(self) -> bool:
        return self.data is None

    def not_empty(self) -> bool:
        return not self.empty()

    def data_encode(self):
        return bencode(self.data)



import logging
import os
import pickle
import re
from datetime import timedelta, datetime
from random import uniform, choices, sample, shuffle, choice

import requests

from notetool.tool.log import log

# 调试日志设置
logger = log('lanzou')
logger.setLevel(logging.INFO)

headers = {
    'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3770.100 Safari/537.36',
    'Referer': 'https://www.lanzous.com',
    'Accept-Language': 'zh-CN,zh;q=0.9',
}


def remove_notes(html: str) -> str:
    """删除网页的注释"""
    # 去掉 html 里面的 // 和 <!-- --> 注释，防止干扰正则匹配提取数据
    # 蓝奏云的前端程序员喜欢改完代码就把原来的代码注释掉,就直接推到生产环境了 =_=
    html = re.sub(r'<!--.+?-->|\s+//\s*.+', '', html)  # html 注释
    html = re.sub(r'(.+?[,;])\s*//.+', r'\1', html)  # js 注释
    return html


def name_format(name: str) -> str:
    """去除非法字符"""
    name = name.replace(u'\xa0', ' ').replace(u'\u3000', ' ').replace('  ', ' ')  # 去除其它字符集的空白符,去除重复空白字符
    return re.sub(r'[$%^!*<>)(+=`\'\"/:;,?]', '', name)


def time_format(time_str: str) -> str:
    """输出格式化时间 %Y-%m-%d"""
    if '秒前' in time_str or '分钟前' in time_str or '小时前' in time_str:
        return datetime.today().strftime('%Y-%m-%d')
    elif '昨天' in time_str:
        return (datetime.today() - timedelta(days=1)).strftime('%Y-%m-%d')
    elif '前天' in time_str:
        return (datetime.today() - timedelta(days=2)).strftime('%Y-%m-%d')
    elif '天前' in time_str:
        days = time_str.replace(' 天前', '')
        return (datetime.today() - timedelta(days=int(days))).strftime('%Y-%m-%d')
    else:
        return time_str


def is_name_valid(filename: str) -> bool:
    """检查文件名是否允许上传"""

    valid_suffix_list = ('ppt', 'xapk', 'ke', 'azw', 'cpk', 'gho', 'dwg', 'db', 'docx', 'deb', 'e', 'ttf', 'xls', 'bat',
                         'crx', 'rpm', 'txf', 'pdf', 'apk', 'ipa', 'txt', 'mobi', 'osk', 'dmg', 'rp', 'osz', 'jar',
                         'ttc', 'z', 'w3x', 'xlsx', 'cetrainer', 'ct', 'rar', 'mp3', 'pptx', 'mobileconfig', 'epub',
                         'imazingapp', 'doc', 'iso', 'img', 'appimage', '7z', 'rplib', 'lolgezi', 'exe', 'azw3', 'zip',
                         'conf', 'tar', 'dll', 'flac', 'xpa', 'lua')

    return filename.split('.')[-1] in valid_suffix_list


def is_file_url(share_url: str) -> bool:
    """判断是否为文件的分享链接"""
    base_pat = r'https?://.+?\.lanzou[six].com/.+'
    user_pat = r'https?://.+?\.lanzou[six].com/i[a-z0-9]{5,}/?'  # 普通用户 URL 规则
    if not re.fullmatch(base_pat, share_url):
        return False
    elif re.fullmatch(user_pat, share_url):
        return True
    else:  # VIP 用户的 URL 很随意
        try:
            html = requests.get(share_url, headers=headers).text
            html = remove_notes(html)
            return True if re.search(r'class="fileinfo"|id="file"|文件描述', html) else False
        except (requests.RequestException, Exception):
            return False


def is_folder_url(share_url: str) -> bool:
    """判断是否为文件夹的分享链接"""
    base_pat = r'https?://.+?\.lanzou[six].com/.+'
    user_pat = r'https?://.+?\.lanzou[six].com/b[a-z0-9]{7,}/?'
    if not re.fullmatch(base_pat, share_url):
        return False
    elif re.fullmatch(user_pat, share_url):
        return True
    else:  # VIP 用户的 URL 很随意
        try:
            html = requests.get(share_url, headers=headers).text
            html = remove_notes(html)
            return True if re.search(r'id="infos"', html) else False
        except (requests.RequestException, Exception):
            return False


def un_serialize(data: bytes):
    """反序列化文件信息数据"""
    try:
        ret = pickle.loads(data)
        if not isinstance(ret, dict):
            return None
        return ret
    except Exception as e:  # 这里可能会丢奇怪的异常
        print(e)
        return None


def big_file_split(file_path: str, max_size: int = 100, start_byte: int = 0) -> (int, str):
    """将大文件拆分为大小、格式随机的数据块, 可指定文件起始字节位置(用于续传)
    :return 数据块文件的大小和绝对路径
    """
    file_name = os.path.basename(file_path)
    file_size = os.path.getsize(file_path)
    tmp_dir = os.path.dirname(file_path) + os.sep + '__' + '.'.join(file_name.split('.')[:-1])

    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)

    def get_random_size() -> int:
        """按权重生成一个不超过 max_size 的文件大小"""
        reduce_size = choices([uniform(0, 20), uniform(20, 30), uniform(30, 60), uniform(60, 80)], weights=[2, 5, 2, 1])
        return round((max_size - reduce_size[0]) * 1048576)

    def get_random_name() -> str:
        """生成一个随机文件名"""
        # 这些格式的文件一般都比较大且不容易触发下载检测
        suffix_list = ('zip', 'rar', 'apk', 'ipa', 'exe', 'pdf', '7z', 'tar', 'deb', 'dmg', 'rpm', 'flac')
        name = list(file_name.replace('.', '').replace(' ', ''))
        name = name + sample('abcdefghijklmnopqrstuvwxyz', 3) + sample('1234567890', 2)
        shuffle(name)  # 打乱顺序
        name = ''.join(name) + '.' + choice(suffix_list)
        return name_format(name)  # 确保随机名合法

    with open(file_path, 'rb') as big_file:
        big_file.seek(start_byte)
        left_size = file_size - start_byte  # 大文件剩余大小
        random_size = get_random_size()
        tmp_file_size = random_size if left_size > random_size else left_size
        tmp_file_path = tmp_dir + os.sep + get_random_name()

        chunk_size = 524288  # 512KB
        left_read_size = tmp_file_size
        with open(tmp_file_path, 'wb') as small_file:
            while left_read_size > 0:
                if left_read_size < chunk_size:  # 不足读取一次
                    small_file.write(big_file.read(left_read_size))
                    break
                # 一次读取一块,防止一次性读取占用内存
                small_file.write(big_file.read(chunk_size))
                left_read_size -= chunk_size

    return tmp_file_size, tmp_file_path


def let_me_upload(file_path):
    """允许文件上传"""
    file_size = os.path.getsize(file_path) / 1024 / 1024  # MB
    file_name = os.path.basename(file_path)

    big_file_suffix = ['zip', 'rar', 'apk', 'ipa', 'exe', 'pdf', '7z', 'tar', 'deb', 'dmg', 'rpm', 'flac']
    small_file_suffix = big_file_suffix + ['doc', 'epub', 'mobi', 'mp3', 'ppt', 'pptx']
    big_file_suffix = choice(big_file_suffix)
    small_file_suffix = choice(small_file_suffix)
    suffix = small_file_suffix if file_size < 30 else big_file_suffix
    new_file_path = '.'.join(file_path.split('.')[:-1]) + '.' + suffix

    with open(new_file_path, 'wb') as out_f:
        # 写入原始文件数据
        with open(file_path, 'rb') as in_f:
            chunk = in_f.read(4096)
            while chunk:
                out_f.write(chunk)
                chunk = in_f.read(4096)
        # 构建文件 "报尾" 保存真实文件名,大小 512 字节
        # 追加数据到文件尾部，并不会影响文件的使用，无需修改即可分享给其他人使用，自己下载时则会去除，确保数据无误
        padding = 512 - len(file_name.encode('utf-8')) - 42  # 序列化后空字典占 42 字节
        data = {'name': file_name, 'padding': b'\x00' * padding}
        data = pickle.dumps(data)
        out_f.write(data)
    return new_file_path


def unit_step(size):
    if size < 1024:
        unit = 'B'
        step = 1
    elif size < 1024 * 1024:
        unit = 'KB'
        step = 1024
    elif size < 1024 * 1024 * 1024:
        unit = 'MB'
        step = 1024 * 1024
    else:
        unit = 'GB'
        step = 1024 * 1024 * 1024
    return unit, step * 1.0



import argparse
import asyncio
import ipaddress
import logging
import os
from pathlib import Path

from aiohttp import web

from . import settings
from .core import Magnet2Torrent
from .dht.network import Server as DHTServer
from .server import routes
from .utils import FailedToFetchException


def main():
    parser = argparse.ArgumentParser(description="Turn a magnet link to torrent.")
    parser.add_argument("--debug", help="Enable debug", action="store_true")
    parser.add_argument("--use-dht", help="Enable DHT", action="store_true", dest="use_dht")
    parser.add_argument("--dht-state-file", help="Where to save DHT info", dest="dht_state_file", type=str, )
    parser.add_argument("--dht-port", help="Port to listen for DHT on", dest="dht_port", type=int,
                        default=settings.DHT_PORT, )
    parser.add_argument("--dht-ip", help="Host to listen for DHT on", dest="dht_ip", type=ipaddress.ip_address,
                        default=ipaddress.IPv4Address("0.0.0.0"), )
    parser.add_argument("--torrent-cache-folder", help="Folder to cache torrent metadata into",
                        dest="torrent_cache_folder", type=str, )
    subparsers = parser.add_subparsers(help="sub-command help", dest="command")

    # dht_test_subparser = subparsers.add_parser("dhttest")

    serve_subparser = subparsers.add_parser(
        "serve", help="Run an HTTP server that serves torrents via an API or directly")
    serve_subparser.add_argument("--ip", type=ipaddress.ip_address, default=ipaddress.IPv4Address("0.0.0.0"),
                                 help="Host to listen on", )
    serve_subparser.add_argument("--port", type=int, default=18667, help="Port to listen on")
    serve_subparser.add_argument("--apikey", type=str, default=None,
                                 help="Protect the endpoint with a simple apikey, add apikey=<apikey> to the url to access", )

    fetch_subparser = subparsers.add_parser("fetch", help="Fetch a torrent and save it locally")
    fetch_subparser.add_argument("magnet", help="Magnet link")

    args = parser.parse_args()

    if args.torrent_cache_folder:
        torrent_cache_folder = Path(args.torrent_cache_folder)

        if not torrent_cache_folder.exists():
            os.makedirs(args.torrent_cache_folder)

        if not torrent_cache_folder.is_dir():
            print("Path {} exists but is not a folder".format(args.torrent_cache_folder))
            quit(1)

    if args.debug:
        logging.basicConfig(level=logging.DEBUG, format="%(asctime)-15s:%(levelname)s:%(name)s:%(lineno)d:%(message)s")

    if args.use_dht:
        print("Bootstrapping DHT server")
        loop = asyncio.get_event_loop()
        dht_server = DHTServer()

        if args.dht_state_file and os.path.isfile(args.dht_state_file):
            dht_server = DHTServer.load_state(args.dht_state_file)
            loop.run_until_complete(dht_server.listen(args.dht_port, str(args.dht_ip)))
        else:
            dht_server = DHTServer()
            loop.run_until_complete(dht_server.listen(args.dht_port, str(args.dht_ip)))
            loop.run_until_complete(dht_server.bootstrap(settings.DHT_BOOTSTRAP_NODES))

        if args.dht_state_file:
            dht_server.save_state_regularly(args.dht_state_file)
        print("Done bootstrapping DHT server")
    else:
        dht_server = None

    if args.command == "serve":
        if not args.debug:
            stdio_handler = logging.StreamHandler()
            stdio_handler.setLevel(logging.INFO)
            logger = logging.getLogger("aiohttp.access")
            logger.setLevel(logging.INFO)
            logger.addHandler(stdio_handler)

        settings.SERVE_APIKEY = args.apikey
        settings.DHT_SERVER = dht_server
        settings.TORRENT_CACHE_FOLDER = args.torrent_cache_folder
        app = web.Application()
        app.add_routes(routes)
        web.run_app(app, host=str(args.ip), port=args.port)
    elif args.command == "fetch":
        loop = asyncio.get_event_loop()
        m2t = Magnet2Torrent(args.magnet, dht_server=dht_server, torrent_cache_folder=args.torrent_cache_folder, )
        try:
            torrent = loop.run_until_complete(m2t.retrieve_torrent())
        except FailedToFetchException:
            print("Unable to fetch magnet link")
            quit(1)
        else:
            with open(torrent.name, "wb") as f:
                f.write(torrent.data_encode())

            print("Downloaded magnet link into file: {}".format(torrent.name))

        if dht_server and args.dht_state_file:
            dht_server.save_state(args.dht_state_file)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



import asyncio
import ipaddress
import logging
import os
from pathlib import Path

from aiohttp import web

from notedrive.magnet import settings
from notedrive.magnet.core import Magnet2Torrent
from notedrive.magnet.dht.network import Server as DHTServer
from notedrive.magnet.server import routes
from notedrive.magnet.utils import FailedToFetchException


def start_serve(dht_server, torrent_cache_folder=None, debug=False, apikey=None, ip=ipaddress.IPv4Address("0.0.0.0"),
                port=18667, ):
    if not debug:
        stdio_handler = logging.StreamHandler()
        stdio_handler.setLevel(logging.INFO)
        logger = logging.getLogger("aiohttp.access")
        logger.setLevel(logging.INFO)
        logger.addHandler(stdio_handler)

    settings.SERVE_APIKEY = apikey
    settings.DHT_SERVER = dht_server
    settings.TORRENT_CACHE_FOLDER = torrent_cache_folder
    app = web.Application()
    app.add_routes(routes)
    web.run_app(app, host=str(ip), port=port)


def start_fetch(dht_server, torrent_cache_folder=None, dht_state_file=None, magnet=None, ):
    loop = asyncio.get_event_loop()
    m2t = Magnet2Torrent(dht_server=dht_server, torrent_cache_folder=torrent_cache_folder, use_additional_trackers=True)
    try:
        torrent = loop.run_until_complete(m2t.retrieve_torrent(magnet_link=magnet))
    except FailedToFetchException:
        print("Unable to fetch magnet link")
        quit(1)
    else:
        with open(torrent.name, "wb") as f:
            f.write(torrent.data_encode())

        print("Downloaded magnet link into file: {}".format(torrent.name))

    if dht_server and dht_state_file:
        dht_server.save_state(dht_state_file)


def start(torrent_cache_folder=None, debug=False, use_dht=False, dht_state_file=None, dht_port=settings.DHT_PORT,
          dht_ip=ipaddress.IPv4Address("0.0.0.0"), command='serve', apikey=None, ip=ipaddress.IPv4Address("0.0.0.0"),
          port=18667, magnet=None, ):
    if torrent_cache_folder:
        torrent_cache_folder = Path(torrent_cache_folder)

        if not torrent_cache_folder.exists():
            os.makedirs(torrent_cache_folder)

        if not torrent_cache_folder.is_dir():
            print("Path {} exists but is not a folder".format(torrent_cache_folder))
            quit(1)

    if debug:
        logging.basicConfig(level=logging.DEBUG, format="%(asctime)-15s:%(levelname)s:%(name)s:%(lineno)d:%(message)s")

    if use_dht:
        print("Bootstrapping DHT server")
        loop = asyncio.get_event_loop()
        dht_server = DHTServer()

        if dht_state_file and os.path.isfile(dht_state_file):
            dht_server = DHTServer.load_state(dht_state_file)
            loop.run_until_complete(dht_server.listen(dht_port, str(dht_ip)))
        else:
            dht_server = DHTServer()
            loop.run_until_complete(dht_server.listen(dht_port, str(dht_ip)))
            loop.run_until_complete(dht_server.bootstrap(settings.DHT_BOOTSTRAP_NODES))

        if dht_state_file:
            dht_server.save_state_regularly(dht_state_file)
        print("Done bootstrapping DHT server")
    else:
        dht_server = None

    if command == "serve":
        start_serve(dht_server, torrent_cache_folder=torrent_cache_folder, debug=debug, apikey=apikey, ip=ip, port=port)
    elif command == "fetch":
        start_fetch(dht_server, torrent_cache_folder=torrent_cache_folder, dht_state_file=dht_state_file, magnet=magnet)



import os

import pycurl

from notedrive.baidu.drive import BaiDuDrive


def download(url, path):
    with open(path, 'wb') as f:
        c = pycurl.Curl()
        c.setopt(pycurl.URL, url)
        c.setopt(pycurl.WRITEDATA, f)
        c.perform()
        c.close()


def download_from_url(url, bduss=None, save_path='/drive/temp/'):
    client = BaiDuDrive(bduss)
    filename = os.path.basename(url)
    download(url, filename)
    client.upload(filename, save_path + filename, overwrite=True)



import asyncio
from itertools import chain
from urllib.parse import urlparse

from notetool.tool.log import log
from .dht.bencode import bdecode
from .dht.torrent import Magnet, Torrent
from .peer import fetch_from_peer
from .peer_tracker import retrieve_peers_http_tracker
from .peer_tracker import retrieve_peers_udp_tracker
from .settings import DEFAULT_TRACKERS

logger = log(__name__)


class Magnet2Torrent:
    def __init__(self, use_trackers=True, use_additional_trackers=False, dht_server=None, torrent_cache_folder=None, ):
        self.use_trackers = use_trackers
        self.use_additional_trackers = use_additional_trackers
        self.dht_server = dht_server
        self.torrent_cache_folder = torrent_cache_folder

    def update_torrent(self, torrent: Torrent, trackers=None) -> Torrent:
        torrent_data = torrent.data
        if self.use_trackers:

            if self.use_additional_trackers:
                trackers += DEFAULT_TRACKERS

            torrent_data[b"announce-list"] = [[tracker.encode("utf-8")] for tracker in trackers]
            if torrent_data[b"announce-list"]:
                torrent_data[b"announce"] = torrent_data[b"announce-list"][0][0]

        return torrent.update_date(torrent_data)

    async def retrieve_torrent(self, magnet_link):
        magnet = Magnet(magnet_link)
        torrent = Torrent(magnet, cache_folder=self.torrent_cache_folder)

        torrent.read_cache()
        if torrent.not_empty():
            self.update_torrent(torrent)
            return torrent

        task_registry = set()
        infohash = magnet.infohash
        tasks = []
        if self.use_trackers:
            trackers = magnet.trackers
            if self.use_additional_trackers:
                trackers += DEFAULT_TRACKERS

            for tracker in trackers:
                try:
                    logger.debug("Trying to fetch peers from {}".format(tracker))
                    tracker_url = urlparse(tracker)
                    if tracker_url.scheme in ["http", "https"]:
                        task = retrieve_peers_http_tracker(task_registry, tracker, infohash)
                    elif tracker_url.scheme in ["udp"]:
                        host, port = tracker_url.netloc.split(":")
                        task = retrieve_peers_udp_tracker(task_registry, host, port, tracker, infohash)
                    else:
                        logger.warn("Unknown scheme, {} {}".format(tracker_url.scheme, tracker_url))
                        continue

                    task = asyncio.ensure_future(task)
                    task.task_type = "tracker"
                    tasks.append(task)
                except Exception as e:
                    logger.error("error {} {}".format(tracker, e))

        if self.dht_server:
            task = asyncio.ensure_future(self.dht_server.find_peers(task_registry, infohash))
            task.task_type = "tracker"
            tasks.append(task)

        handled_peers = set()
        while tasks:
            done, tasks = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
            for task in done:
                result = task.result()
                if task.task_type == "tracker":
                    for peer in result[1]["peers"]:
                        if peer in handled_peers:
                            continue
                        handled_peers.add(peer)
                        peer_ip, peer_port = peer
                        logger.debug("Connecting to {}:{}".format(peer_ip, peer_port))
                        peer_task = asyncio.ensure_future(fetch_from_peer(task_registry, peer_ip, peer_port, infohash))
                        peer_task.task_type = "peer"
                        tasks.add(peer_task)
                    if len(result) > 2:
                        new_task = asyncio.ensure_future(result[2]())
                        new_task.task_type = task.task_type
                        tasks.add(new_task)
                elif task.task_type == "peer":
                    if result:
                        for task1 in chain(task_registry, tasks):
                            if not task1.done():
                                task1.cancel()

                        torrent.from_data({b"info": bdecode(result)}, decode=False)
                        self.update_torrent(torrent, trackers=magnet.trackers)
                        torrent.save_cache()
                        return torrent

        raise Exception('FailedToFetchException')

    def get_magnet_info(self, magnet_link: str):

        try:
            # torrent = asyncio.run(self.retrieve_torrent(magnet_link))
            # torrent = asyncio.run_coroutine_threadsafe(self.retrieve_torrent(magnet_link))
            torrent = asyncio.get_event_loop().run_until_complete(self.retrieve_torrent(magnet_link))
            return torrent
        except Exception as e:
            logger.error("Failed {}".format(e))
            return None



import base64
from urllib.parse import quote

from aiohttp import web

from . import settings
from .core import Magnet2Torrent
from .utils import FailedToFetchException

routes = web.RouteTableDef()


@routes.get("/")
async def get_torrent(request):
    if settings.SERVE_APIKEY and request.query.getone("apikey", None) != settings.SERVE_APIKEY:
        raise web.HTTPUnauthorized()

    try:
        magnet = request.query["magnet"]
    except:
        return web.json_response({"status": "error", "message": "magnet argument missing from url"}, status=400, )

    m2t = Magnet2Torrent(magnet, dht_server=settings.DHT_SERVER, torrent_cache_folder=settings.TORRENT_CACHE_FOLDER, )
    try:
        torrent = await m2t.retrieve_torrent()
    except FailedToFetchException:
        return web.json_response({"status": "error", "message": "failed to retrieve magnet link"}, status=500)

    if request.query.getone("direct", None) is not None:
        return web.Response(body=torrent.data_encode(),
                            headers={"Content-Disposition": "attachment; filename*=UTF-8''{}".format(
                                quote(torrent.name))},
                            )
    else:
        return web.json_response(
            {"status": "success", "filename": torrent.name,
             "torrent_data": base64.b64encode(torrent.data_encode()).decode("utf-8")}
        )



import concurrent.futures
import os
from threading import Lock
from time import sleep

import numpy as np
import pycurl
from requests import Session
from tqdm import tqdm

from notetool.tool import exists_file, delete_file, path_parse
from notetool.tool import log


class BaseDownLoad:
    def __init__(self, session=None):
        self.url = None
        self.path = None
        self.size = None
        self.session = session or Session()
        self.logger = log(__name__)
        self.logger.debug(self.session.headers)

    def download(self, url, path, size=0, overwrite=False):
        path = path_parse(path)
        self.url = url
        self.path = path
        if exists_file(file_path=path, mkdir=True):
            if overwrite:
                delete_file(file_path=path)
            else:
                self.logger.info('file exist and return[path={}].'.format(path))
                return
        else:
            self.init_path(path)

        file_name = os.path.basename(path)
        self.logger.info("download {} from {} to {} ".format(file_name, url, path))
        self._download(url, path, size=size)
        self.logger.info('download {} success'.format(file_name))

    def _download(self, url, path, size=0):
        pass

    @staticmethod
    def init_path(path):
        local_dir = os.path.dirname(path)
        if not os.path.exists(local_dir):
            os.makedirs(local_dir)


class MultiThreadDownload(BaseDownLoad):
    def __init__(self, worker=5, chunk_size=10 * 1024 * 1024, *args, **kwargs):
        self.lock = Lock()
        self.num = worker
        self.worker = worker
        self.chunk_size = chunk_size
        self.down_btn = None
        super(MultiThreadDownload, self).__init__(*args, **kwargs)

    def down(self, start, end):
        headers = {'Range': 'bytes={}-{}'.format(start, end)}
        # stream = True 下载的数据不会保存在内存中
        r = self.session.get(self.url, headers=headers)
        # 写入文件对应位置,加入文件锁
        self.lock.acquire()
        print(r.status_code)

        with open(self.path, "rb+") as fp:
            fp.seek(start)
            fp.write(r.content)
            self.lock.release()
            # 释放锁
        self.down_btn.update(1)

    def _download(self, url, path, size=0, **kwargs):
        if size == 0:
            r = self.session.head(self.url)

            # 若资源显示302,则迭代找寻源文件
            # while r.status_code == 302:
            #     self.url = r.headers['Location']
            #     logger.warning("该url已重定向至{}".format(self.url))
            #     r = self.session.head(self.url)

            self.size = int(r.headers['Content-Length'])
        else:
            self.size = size

        self.logger.info('该文件大小为：{} bytes'.format(self.size))
        self.num = self.size // self.chunk_size + 1

        self.down_btn = tqdm(total=self.num)
        # 创建一个和要下载文件一样大小的文件
        with open(self.path, "wb") as fp:
            fp.truncate(self.size)

        pool = concurrent.futures.ThreadPoolExecutor(max_workers=self.worker)

        # 启动多线程写文件
        futures = []

        for i in range(self.num):
            start = self.chunk_size * i
            # 最后一块
            if i == self.num - 1:
                end = self.size
            else:
                end = start + self.chunk_size + 1
            sleep(0.5)
            futures.append(pool.submit(self.down, start, end))
        concurrent.futures.wait(futures)


class SingleThreadDownload(BaseDownLoad):
    """
    HTTP下载较大文件的工具
    """

    def __init__(self, chunk=1024 * 1024, *args, **kwargs):
        self.position = 0  # 当前的字节偏移量
        self.chunk = chunk

        super(SingleThreadDownload, self).__init__(*args, **kwargs)

    def _download(self, url, path, size=0):
        file_size = size
        local_path = path
        file_name = os.path.basename(local_path)

        if os.path.exists(local_path):
            first_byte = os.path.getsize(local_path)
        else:
            first_byte = 0

        self.init_path(local_path)

        position = first_byte

        mode = 'ab' if first_byte > 0 else 'wb'
        with open(local_path, mode=mode) as f:
            with tqdm(initial=np.round(first_byte / self.chunk, 2),
                      total=np.round(file_size / self.chunk, 2),
                      unit='MB',
                      desc=file_name) as pbar:
                while position < file_size:
                    interval = (position, position + self.chunk)
                    position += (self.chunk + 1)

                    time = 3
                    while time > 0:
                        try:
                            resp = self.session.get(url, headers={'Range': 'bytes=%s-%s' % interval})
                            f.write(resp.content)
                            pbar.update(1)
                            break
                        except Exception as e:
                            time -= 1
                            self.logger.warning(e)

        return True


class PyCurlDownLoad(BaseDownLoad):
    def __init__(self, *args, **kwargs):
        super(PyCurlDownLoad, self).__init__(*args, **kwargs)

    def _download(self, url, path, size=0):
        with open(self.path, 'wb') as f:
            c = pycurl.Curl()
            c.setopt(pycurl.URL, self.url)
            c.setopt(pycurl.WRITEDATA, f)
            c.perform()
            c.close()


def download(url, path, size=0, session=None, overwrite=False, mode='curl'):
    if mode == 'curl':
        down = PyCurlDownLoad()
    elif mode == 'single':
        down = SingleThreadDownload(session=session)
    else:
        down = MultiThreadDownload(session=session)
    down.download(url, path, size, overwrite=overwrite)
